(* ::Package:: *)

(*:Package Version: 1.02 *)

(*:Version History: 1.01 (2014.05.04 - 2014.05.18)BoettcherPhi
                    1.02 (2014.05.19 -           )SmoothIteration *)

(*:Copyright: Copyright 2014.05-2014.05, Souichiro Ikebe *)

(*:Author: Souichiro Ikebe *)

(*:Name: SpecialFunctions`FractalRelated` *)

(*:Title: BoettcherPhi *)

(*:Keywords: JuliaSetQ, MandelbrotSetQ, JuliaSetDistance, MandelbrotSetDistance, 
             JuliaBoettcherPhi, MandelbrotBoettcherPhi *)

(*:Summary:
\:3000\:3000This package enable calculation of the function that is related to fractal. 
   Specifically, Boettcher-functions of the Julia-set, Boettcher-functions of 
   the Mandelbrot-set etc. are applicable.
*)

(**********************************************************************************)





BeginPackage["SpecialFunctions`FractalRelated`"]





JuliaSetQ::usage =
"JuliaSetQ[c,z,iMax] returns True if z is in the Julia-set, and False otherwise.
 Here, iMax is number of iterations of the complex mapping.
 In addition, JuliaSetQ[c,z] is the same as JuliaSetQ[c,z,100]."

MandelbrotSetQ::usage =
"MandelbrotSetQ[c,iMax] returns True if c is in the Mandelbrot-set, 
 and False otherwise.
 Here, iMax is number of iterations of the complex mapping.
 In addition, MandelbrotSetQ[c] is the same as MandelbrotSetQ[c,100]."

JuliaSetDistance::usage =
"JuliaSetDistance[c,z,iMax] gives distance from z to the Julia-set.
 Here, iMax is number of iterations of the complex mapping.
 In addition, JuliaSetDistance[c,z] is the same as JuliaSetDistance[c,z,100]."

MandelbrotSetDistance::usage =
"MandelbrotSetDistance[c,iMax] gives distance from z to the Mandelbrot-set.
 Here, iMax is number of iterations of the complex mapping.
 In addition, MandelbrotSetDistance[c] is the same as 
 MandelbrotSetDistance[c,100]."

JuliaBoettcherPhi::usage =
"JuliaBoettcherPhi[c,z,iMax] gives Boettcher-functions of the Julia-set.
 Here, iMax is number of iterations of the complex mapping.
 In addition, JuliaBoettcherPhi[c,z] is the same as JuliaBoettcherPhi[c,z,50]."

MandelbrotBoettcherPhi::usage =
"MandelbrotBoettcherPhi[c,iMax] gives Boettcher-functions of the Mandelbrot-set.
 Here, iMax is number of iterations of the complex mapping.
 In addition, MandelbrotBoettcherPhi[c] is the same as 
 MandelbrotBoettcherPhi[c,100]."





Begin["SpecialFunctions`FractalRelated`Private`"]





(**********************************************************************************)
(*                              JuliaSetQ[c,z,iMax]                               *)
(**********************************************************************************)

JuliaSetQ[c_,z_,iMax_Integer:100]:=
  innerJuliaSetQ[c,z,iMax]/;
  (Precision[z]<Infinity && iMax>=0)

innerJuliaSetQ=
  Compile[{{c,_Complex},{z,_Complex},{iMax,_Integer}},
    Module[{i=0,f=z},
      While[i<iMax && Abs[f]<=2,f=f^2+c;i++];
    If[Abs[f]<=2,True,False]]];



(**********************************************************************************)
(*                             MandelbrotSetQ[c,iMax]                             *)
(**********************************************************************************)

MandelbrotSetQ[c_,iMax_Integer:100]:=
  innerMandelbrotSetQ[c,iMax]/;
  (Precision[c]<Infinity && iMax>=0)

innerMandelbrotSetQ=
  Compile[{{c,_Complex},{iMax,_Integer}},
    Module[{i=0,f=c},
      While[i<iMax && Abs[f]<=2,f=f^2+c;i++];
    If[Abs[f]<=2,True,False]]];



(**********************************************************************************)
(*                           JuliaSetDistance[c,z,iMax]                           *)
(**********************************************************************************)

JuliaSetDistance[c_,z_,iMax_Integer:100]:=
  If[Not[innerJuliaSetQ[c,z,iMax]],distanceJulia[c,z,iMax]]/;
  (Precision[z]<Infinity && iMax>=0)



(**********************************************************************************)
(*                         MandelbrotSetDistance[c,iMax]                          *)
(**********************************************************************************)

MandelbrotSetDistance[c_,iMax_Integer:100]:=
  If[Not[innerMandelbrotSetQ[c,iMax]],distanceMandel[c,iMax]]/;
  (Precision[c]<Infinity && iMax>=0)



(**********************************************************************************)
(*                          JuliaBoettcherPhi[c,z,iMax]                           *)
(**********************************************************************************)

JuliaBoettcherPhi[c_,z_,iMax_Integer:50]:=
  If[Not[innerJuliaSetQ[c,z,iMax]],
    Exp[AbsJuliaBoettcherPhi[c,z,iMax]+I*ArgJuliaBoettcherPhi[c,z,iMax]]]/;
  (Precision[z]<Infinity && iMax>=0)

AbsJuliaBoettcherPhi=
  Compile[{{c,_Complex},{z,_Complex},{iMax,_Integer}},
    Module[{s=0.+0*I,so=-1.+0*I,i=0,f=z,b=2.},
      While[(i<iMax && so!=s),so=s;
        s+=Log[Abs[1+c/(f^2)]]/b;
        f=f^2+c;b*=2;i++];
      If[Abs[f]>2,s+Log[Abs[z]]]]];

ArgJuliaBoettcherPhi=
  Compile[{{c,_Complex},{z,_Complex},{iMax,_Integer}},
    Module[{s=0.+0*I,so=-1.+0*I,i=0,f=z,b=2.,smplQ},
      smplQ=If[Abs[((z^2+c)^2+c)^2+c]>2,True,False];
      While[(i<iMax && so!=s),so=s;
        s+=If[smplQ,Arg[1+c/(f^2)],jBoettcherArgTerm[c,z,i,iMax]]/b;
        f=f^2+c;b*=2;i++];
      If[Abs[f]>2,s+Arg[z]]],
    CompilationOptions->{"InlineExternalDefinitions"->True}];

(*---------  Series terms of the argument Julia-Boettcher function  ---------*)
jBoettcherArgTerm[c_,z_,i_,iMax_]:=
  Module[{argT},
    argT=Re[argTermJuliaB[c,z,i]];
    If[Im[c]>=0,
      If[-0.5<=argT<=Pi,argT,If[branchChangeJQ[c,z,i,iMax],argT+2 Pi,argT]],
      If[-Pi<argT<=0.5,argT,If[branchChangeJQ[c,z,i,iMax],argT-2 Pi,argT]]]]/;
  (Precision[z]<Infinity)

(*--------------------  Test of the Branch point Change  --------------------*)
branchChangeJQ=
  Compile[{{c,_Complex},{z,_Complex},{i,_Integer},{iMax,_Integer}},
    Module[{j=1,z1,d,pointL,ordL,argT},
      z1=z;
      d=distanceJulia[c,z,iMax];
      argT=Re[argTermJuliaB[c,z,i]];
      While[j<iMax && If[Im[c]>=0,-Pi<=argT<-0.5,0.5<argT<=Pi],
        pointL=z1+d*Exp[Pi*I/8.0*Range[-7,8]];
        ordL=Ordering[Map[juliaSmoothIter[c,#,iMax]&,pointL],1];
        z1=Part[pointL,Flatten[{ordL}][[1]]];
        d=distanceJulia[c,z1,iMax];
        argT=Re[argTermJuliaB[c,z1,i]];j++];
      If[If[Im[c]>=0,Pi/2<argT<=Pi,-Pi<=argT<-Pi/2],
        True,False]],
    CompilationOptions->{"InlineExternalDefinitions"->True}];

(*-------------------------  Code of the sub loops  -------------------------*)
distanceJulia=
  Compile[{{c,_Complex},{z,_Complex},{iMax,_Integer}},
    Module[{i=0,f=0.+0*I,f0=z,df=1.+0*I,ans=0.,ansO=-1.},
      While[i<iMax && ans!=ansO,
        ansO=ans;
        f=f0^2+c;df=2*f0*df;
        ans=Abs[f]/Abs[df]*Log[Abs[f]];
        f0=f;i++];ans]];

juliaSmoothIter=
  Compile[{{c,_Complex},{z,_Complex},{iMax,_Integer}},
    Module[{i=0,f=z,r=6,cet},
      While[i<iMax && Abs[f]<=r^2,
        f=f^2+c;i++];
      cet=If[i!=iMax,i+(Log[Log[r]]-Log[Log[Abs[f]]])/Log[2],i];
      Min[cet/iMax,1]]];

argTermJuliaB=
  Compile[{{c,_Complex},{z,_Complex},{i,_Integer}},
    Module[{j=0,f=z},
      While[j<i,f=f^2+c;j++];
        Arg[1+c/(f^2)]]];



(**********************************************************************************)
(*                         MandelbrotBoettcherPhi[c,iMax]                         *)
(**********************************************************************************)

MandelbrotBoettcherPhi[c_,iMax_Integer:100]:=
  If[Not[innerMandelbrotSetQ[c,iMax]],
    Exp[AbsMandelbrotBoettcherPhi[c,iMax]+I*ArgMandelbrotBoettcherPhi[c,iMax]]]/;
  (Precision[c]<Infinity && iMax>=0)

AbsMandelbrotBoettcherPhi=
  Compile[{{c,_Complex},{iMax,_Integer}},
    Module[{s=0.+0*I,so=-1.+0*I,i=0,f=c,b=2.0},
      While[(i<iMax && so!=s),so=s;
        s+=Log[Abs[1+c/(f^2)]]/b;
        f=f^2+c;b*=2;i++];
    If[Abs[f]>2,s+Log[Abs[c]]]]];

ArgMandelbrotBoettcherPhi=
  Compile[{{c,_Complex},{iMax,_Integer}},
    Module[{s=0.+0*I,so=-1.+0*I,i=0,f=c,b=2.0,smplQ},
      smplQ=If[Abs[((c^2+c)^2+c)^2+c]>2,True,False];
      While[(i<iMax && so!=s),so=s;
        s+=If[smplQ,Arg[1+c/(f^2)],mBoettcherArgTerm[c,i,iMax]]/b;
        f=f^2+c;b*=2;i++];
    If[Abs[f]>2,s+Arg[c]]],
    CompilationOptions->{"InlineExternalDefinitions"->True}];

(*------  Series terms of the argument Mandelbrot-Boettcher function  -------*)
mBoettcherArgTerm[c_,i_,iMax_]:=
  Module[{argT},
    argT=Re[argTermMandelB[c,i]];
    If[Im[c]>=0,
      If[-1.6<=argT<=Pi,argT,If[branchChangeMQ[c,i,iMax],argT+2 Pi,argT]],
      If[-Pi<argT<=1.6,argT,If[branchChangeMQ[c,i,iMax],argT-2 Pi,argT]]]]/;
  (Precision[c]<Infinity)

(*--------------------  Test of the Branch point Change  --------------------*)
branchChangeMQ=
  Compile[{{c,_Complex},{i,_Integer},{iMax,_Integer}},
    Module[{j=1,c1,d,pointL,ordL,argT},
      c1=c;
      d=distanceMandel[c,iMax];
      argT=Re[argTermMandelB[c,i]];
      While[j<iMax && If[Im[c]>=0,-Pi<=argT<-1.6,1.6<argT<=Pi],
        pointL=c1+d*Exp[Pi*I/8.0*Range[-7,8]];
        ordL=Ordering[Map[mandelbrotSmoothIter[#,iMax]&,pointL],1];
        c1=Part[pointL,Flatten[{ordL}][[1]]];
        d=distanceMandel[c1,iMax];
        argT=Re[argTermMandelB[c1,i]];j++];
      If[If[Im[c]>=0,Pi/2<argT<=Pi,-Pi<=argT<-Pi/2],
        True,False]],
    CompilationOptions->{"InlineExternalDefinitions"->True}];

(*-------------------------  Code of the sub loops  -------------------------*)
distanceMandel=
  Compile[{{c,_Complex},{iMax,_Integer}},
    Module[{i=0,f=0.+0*I,f0=c,df=1.+0*I,ans=0.+0*I,ansO=-1.+0*I},
      While[i<iMax && ans!=ansO,
        ansO=ans;
        f=f0^2+c;df=2*f0*df;
        ans=Abs[f]/Abs[df]*Log[Abs[f]];
        f0=f;i++];ans]];

mandelbrotSmoothIter=
  Compile[{{c,_Complex},{iMax,_Integer}},
    Module[{i=0,f=c,r=6,cet},
      While[i<iMax && Abs[f]<=r^2,
        f=f^2+c;i++];
      cet=If[i!=iMax,i+(Log[Log[r]]-Log[Log[Abs[f]]])/Log[2],i];
      Min[cet/iMax,1]]];

argTermMandelB=
  Compile[{{c,_Complex},{i,_Integer}},
    Module[{j=0,f=c},
      While[j<i,f=f^2+c;j++];
        Arg[1+c/(f^2)]]];



(**********************************************************************************)

End[]

Protect[JuliaSetQ, MandelbrotSetQ, JuliaSetDistance, MandelbrotSetDistance, 
        JuliaBoettcherPhi, MandelbrotBoettcherPhi]

EndPackage[]
